/*
 * Author: Jos Visser (josv@osp.nl)
 * 
 * Change log:
 * ------------------------------------------------------------
 * 17may2000	josv@osp.nl	Lifted restriction on not having spaces
 *				in the base DN.
 * 04dec2000	josv@osp.nl	Changed to comparison for the object
 *				class to be case insensitive.
 *				Netscape 4.x directory server users will
 *				be happy about this.
 * 25jan2001	josv@osp.nl	Changes for WU FTPD 2.6.1 and OpenLDAP 2.0.7
 *
 * The following is the "BSD" License for this software:
 *
 * -----------------------------------------------------------------------
 * Copyright (c) 1999-2001, Jos Visser <josv@osp.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the distribution.
 * 
 * Neither name of Open Solution Providers, nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission. 

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. 
 * -----------------------------------------------------------------------
 *
 * This file is the heart of the WU-FTP LDAP feature that I implemented
 * because I felt that I needed it. Please feel free to adopt, adapt and
 * improve this source, but send me (josv@osp.nl) a note when you do.
 *
 * For more elaborate documentation, see README.LDAP.
 */

#include "config.h"

#ifdef USE_LDAP

#include <unistd.h>

#include "portable.h"

#include <stdlib.h>

#include <ac/string.h>
#include <ac/unistd.h>

#include "lutil_md5.h"
#include "lutil_sha1.h"
#include "lutil.h"

#include <lber.h>
#include <ldap.h>

#include <sys/time.h>
#include <pwd.h>
#include <syslog.h>

#include "extensions.h"

/*
 * A user found in the directory must be of this objectClass to be
 * deemed a valid FTP user.
 */
#define FTP_OBJECTCLASS 	"ftpUser"

#define FREE(x)			{ if ((x)) free((x)); (x)=0; }

void reply(int code, char *fmt, ...);
void lreply(int code, char *fmt, ...);

/*
 * It seems that global variables are coming in vogue again.
 */
static LDAP *ld=0;
static struct timeval tv;
static struct passwd pw;
static char *dn=0;
static char *cn=0;
static char base64digest[29];
static char *diruser=0;
static char *dirpass=0;
static char *dirhost="localhost";
static int dirport=LDAP_PORT;
static char *dirbase="o=osp.nl";

/* 
 * The list of entry attributes that I am interested in. An ldap_search
 * will return only these attributes.
 */
static char *ftpattrs[] = {
	"objectClass",
	"cn",
	"gidNumber",
	"uidNumber",
	"homeDirectory",
	"gecos",
	0
};

/*
 * Something went wrong. Let's try to give a meaningful error. The error
 * is written to the syslog, and to the FTP user (at the client side) as
 * a 500 permanent error reply. The FTP client on the remote site will
 * terminate the connection.
 */
static void *error(int err, char *s)
{
	char *msg;

	if (err>=0) {
		msg=ldap_err2string(err);
		syslog(LOG_ERR,"LDAP error: %s reports %s",s,msg);
		reply(500,"LDAP error: %s reports %s",s,msg);
	} else {
		syslog(LOG_ERR,"General LDAP error: %s",s);
		reply(500,"General LDAP error: %s",s);
	}

	return 0;
}

/*
 * Get a parameter from the ftpaccess configuration file. This routine
 * automatically concatenates all parameters on the ftpaccess line. This
 * is used for entries (like the bind DN or the base DN) that can
 * contain spaces.
 */
static char *my_getaclentry(char *key)
{
	struct aclmember *entry=0;
	int i,len;
	char *buf;

	if (getaclentry(key, &entry) && ARG0) {

		for (len=i=0; i<MAXARGS && ARG[i]; i++)
			len+=strlen(ARG[i])+1;

		buf=malloc(len+1);
		*buf=0;

		for (i=0; i<MAXARGS && ARG[i]; i++) {
			if (i>0) strcat(buf," ");
			strcat(buf,ARG[i]);
		}

		return buf;
	}

	return 0;
}

/*
 * Initialise the connection to the directory server. First retrieve the
 * necessary parameters from the ftpaccess file, the set up an LDAP
 * connection.
 */
static int init_ldap() 
{
	struct aclmember 	*entry;
	int			rc;

	entry=0;
	if (getaclentry("ldap-host", &entry) && ARG0) dirhost=ARG0;

	entry=0;
	if (getaclentry("ldap-port", &entry) && ARG0) dirport=atoi(ARG0);

	/*
	 * An LDAP bind dn can (and almost always will) contain spaces.
	 * Therefore, I construct the bind dn from the entire set of
	 * arguments specified on the "ldap-user" line.
	 */
	FREE(diruser);
	diruser=my_getaclentry("ldap-user");

	/*
	 * Currently, the password is specified in clear text in the
	 * ftpaccess file. I must do something about that.
	 */
	entry=0;
	if (getaclentry("ldap-password", &entry) && ARG0) dirpass=ARG0;

	/*
	 * Read the bind DN from the ftpaccess file. Since this
	 * parameter can contain spaces, the my_getaclentry() function
	 * is used.
	 */
	dirbase=my_getaclentry("ldap-base");

	/*
	 * syslog(LOG_INFO,"LDAP parameters:");
	 * syslog(LOG_INFO,"server   = %s",dirhost);
	 * syslog(LOG_INFO,"port     = %d",dirport);
	 * syslog(LOG_INFO,"base dn  = %s",dirbase);
	 * syslog(LOG_INFO,"bind dn  = %s",diruser?diruser:"(anonymous)");
	 * syslog(LOG_INFO,"password = %s",dirpass?"(withheld)":"(none)");
	 */

	/*
	 * Build a connection to the directory server.
	 */
	if (!(ld=ldap_init(dirhost,dirport)))
		return (int)error(0,"ldap_init()");

	/*
	 * Bind to the directory server as the specified user (bind dn)
	 * and with the specified password. If they are not specified in
	 * the ftpaccess file, we bind as the anonymous user, but it is
	 * then up to you to make sure that this user has enough rights
	 * to perform searches and compares in the directory. Especially
	 * the userPassword attribute is critical, since most
	 * directories protect this attribute from anonymous access by
	 * default.
	 */
	if ((rc=ldap_simple_bind_s(ld,diruser,dirpass))!=LDAP_SUCCESS) 
		return (int)error(rc,"ldap_bind");

	/*
	 * This struct timeval is used to limit the maximum time an LDAP
	 * search operation can take to 10 seconds.
	 */
	tv.tv_sec=10;
	tv.tv_usec=0;

	return 1;
}

/*
 * Check whether the user we found in the directory is of the correct
 * objectClass. I could have fired off an ldap_compare_s(), but since we
 * have the data in storage on the client, why not go through the data
 * myself?
 */
static int check_ftpUser(LDAP *ld, LDAPMessage *entry)
{
	char **vals;
	int i;

	vals=ldap_get_values(ld,entry,"objectClass");

	if (!vals) return 0;

	for (i=0; vals[i]; i++) 
		if (strcasecmp(vals[i],FTP_OBJECTCLASS)==0) {
			ldap_value_free(vals);
			return 1;
		}

	ldap_value_free(vals);
	return 0;
}

/* 
 * Return an LDAP entry attribute of the user as a string.
 */
static char *attr_str(LDAP *ld, LDAPMessage *entry, char *attr, char *def)
{
	char **vals;
	char *result;

	vals=ldap_get_values(ld,entry,attr);

	if (vals) {
		result=strdup(vals[0]);
		ldap_value_free(vals);
	} else if (def)
		result=strdup(def);
	else
		result=0;

	return result;
}

/* 
 * Return an LDAP entry attribute of the user as an integer.
 */
static int attr_num(LDAP *ld, LDAPMessage *entry, char *attr, int def)
{
	char **vals;
	int result;

	vals=ldap_get_values(ld,entry,attr);

	if (vals) {
		result=atoi(vals[0]);
		ldap_value_free(vals);
	} else
		result=def;

	return result;
}

/*
 * Search the directory for the given user, and fill an industry
 * standard "struct passwd" with the values found. The main of the WU
 * FTP daemon has got no clue where the data came from, and will use it
 * regardless. Returns 0 if the * user can not be found, or is invalid.
 */
struct passwd *getpwnam_ldap(char *name)
{
	int 		n;
	LDAPMessage 	*result,*entry;
	char		buf[80];

	/*
	 * Be liberal in what you accept, and strict in what you emit.
	 */
	if (!name) return 0;	/* No user name specified */
	if (*name==0) return 0; /* Empty user name specified */

	/*
	 * Initialise the LDAP connection.
	 */
	if (!ld) if (!init_ldap()) return 0;

	/*
	 * Build the LDAP search filter. Rather simple, but exactly what
	 * we need.
	 */
	sprintf(buf,"(uid=%s)",name);

	/* 
	 * Fire off the search operation. Only retrieves the attributes
	 * I'm interested in.
	 */
	n=ldap_search_st(ld,dirbase,LDAP_SCOPE_SUBTREE,buf,
		ftpattrs,0,&tv,&result);

	/*
	 * Oh, oh, the user does not exist!
	 */
	if (n==LDAP_NO_SUCH_OBJECT) return 0;

	/*
	 * "If we don't succeed, we run the risk of failure" -- D. Quale
	 */
	if (n!=LDAP_SUCCESS) return error(n,"search");

	/*
	 * Count how much entries we got back from the directory.
	 */
	if ((n=ldap_count_entries(ld,result))==-1) {
		n=ldap_result2error(ld,result,1);
		return error(n,"count");
	}

	/*
	 * Hmm, we got back more than one! I'll use the first one but
	 * let's be nice and give a warning.
	 */
	if (n>1) syslog(LOG_WARNING,"ldap_search returned %d entries for %s",
			n,buf);

	/*
	 * Hmm, we got back nothing. Shouldn't this have given a 
	 * NO_SUCH_OBJECT error?
	 */
	if (n==0) return 0;

	/*
	 * Retrieve the first entry of the result set
	 */
	entry=ldap_first_entry(ld,result);

	if (!entry) {
		syslog(LOG_ERR,"ldap_first_entry return NULL!");
		return 0;
	}

	/*
	 * Retrieve the Distinguised Name (DN) and Common Name (CN) of
	 * the user. And log who tries to log on.
	 */
	FREE(dn);
	dn=ldap_get_dn(ld,entry);
	FREE(cn);
	cn=attr_str(ld,entry,"cn",0);
	syslog(LOG_INFO,"%s recognised as %s (%s)",name,dn,cn?cn:"");

	/*
	 * Check whether this user is of the correct objectClass
	 * (ftpUser). If not, let's pretend he does not exist at all.
	 */
	if (!check_ftpUser(ld,entry)) {
		syslog(LOG_ERR,"%s is not of type (objectclass) %s",
			name,FTP_OBJECTCLASS);
		return 0;
	}

	/*
	 * Ok, I now fill the struct passwd with the fields from the
	 * directory entry.
	 */
	FREE(pw.pw_name);
	FREE(pw.pw_passwd);
	FREE(pw.pw_gecos);
	FREE(pw.pw_dir);
	FREE(pw.pw_shell);

	pw.pw_name=strdup(name);
	pw.pw_passwd=strdup("secret :-)");
	pw.pw_gecos=attr_str(ld,entry,"gecos","");
	pw.pw_dir=attr_str(ld,entry,"homeDirectory",0);

	/*
	 * This should not happen, but for those who have not configured
	 * the directory correctly... (mind you, schema checking *is* a
	 * good idea).
	 */
	if (!pw.pw_dir) {
		syslog(LOG_ERR,"%s has no homeDirectory attribute",name);
		return 0;
	}

	/*
	 * An empty shell is handled correctly by the core of WU FTPD.
	 */
	pw.pw_shell=strdup("");
	pw.pw_uid=attr_num(ld,entry,"uidNumber",-1);

	/*
	 * Again, this should not happen.
	 */
	if (pw.pw_uid<0) {
		syslog(LOG_ERR,"%s has no uidNumber attribute",name);
		return 0;
	}

	pw.pw_gid=attr_num(ld,entry,"gidNumber",-1);

	/*
	 * And again, this should not happen as well.
	 */
	if (pw.pw_gid<0) {
		syslog(LOG_ERR,"%s has no gidNumber attribute",name);
		return 0;
	}

	/*
	 * Ok, we're done. Bring out the dancing girls!
	 */
	ldap_msgfree(result);

	return &pw;
}

/*
 * Close the LDAP connection. Quiz question: from where is this called?
 */
void close_ldap()
{
	if (!ld) return;

	ldap_unbind(ld);
	ld=0;
}

/*
 * Calculate the MD5 hash of a password. This code has been adapted from
 * the OpenLDAP passwd.c source in liblutil. Thanks guys!
 */
static char *md5_passwd(char *cred)
{
	lutil_MD5_CTX MD5context;
	unsigned char MD5digest[16];

	lutil_MD5Init(&MD5context);
	lutil_MD5Update(&MD5context,(const unsigned char *)cred,strlen(cred));
	lutil_MD5Final(MD5digest, &MD5context);

	if (lutil_b64_ntop(MD5digest, sizeof(MD5digest), base64digest, 
					sizeof(base64digest)) < 0)
			return 0;

	return base64digest;
}

/*
 * Calculate the SHA hash of a password. This code has been adapted from the
 * OpenLDAP passwd.c source in liblutil. Thanks guys!
 */
static char *sha_passwd(char *cred)
{
	lutil_SHA1_CTX SHA1context;
	unsigned char SHA1digest[20];

	lutil_SHA1Init(&SHA1context);
	lutil_SHA1Update(&SHA1context,(const unsigned char *)cred,strlen(cred));
	lutil_SHA1Final(SHA1digest, &SHA1context);

	if (lutil_b64_ntop(SHA1digest, sizeof(SHA1digest),
			base64digest, sizeof(base64digest)) < 0)
		return 0;

	return base64digest;
}

/* 
 * This helper routine checkes whether its argument string happens to be
 * the same as the content of the userPassword field of the user trying
 * to log on. For obvious reasons, this is done with an ldap_compare(),
 * so it is the server that does the comparison. Any hashing or crypting
 * of the password has already been done by the caller.
 */
static int check_pass(char *s)
{
	int n=ldap_compare_s(ld,dn,"userPassword",s);

	if (n==LDAP_COMPARE_TRUE)	return 1;
	if (n==LDAP_COMPARE_FALSE)	return 0;

	error(n,"ldap_compare");
	return 0;
}

/*
 * This routine is the core of the password checking routine. It takes
 * the password supplied by the FTP user and tries to match it with the
 * userPassword field of the user in the directory (through the
 * check_pass helper). Most directories can hash the userPassword so
 * that it is obscured from the directory administrator. This routine
 * tries the most common hashing algorithms: SHA and MD5. If they do not
 * match, a clear text match is tried.
 */
static int check_pass_ldap_horse(char *passwd)
{
	char *s;
	char buf[80];

	/*
	 * This should not happen.
	 */
	if (!ld || !dn) return 0;

	/*
	 * Hash the supplied password with SHA and compare it.
	 */
	s=sha_passwd(passwd);

	if (s) {
		sprintf(buf,"{SHA}%s",s);
		if (check_pass(buf)) return 1;
	}

	/*
	 * Hash the supplied password with MD5 and compare it.
	 */
	s=md5_passwd(passwd);

	if (s) {
		sprintf(buf,"{MD5}%s",s);
		if (check_pass(buf)) return 1;
	}

	/*
	 * If all else fails, try a clear text match, but give a warning
	 * if it matches. You better configure your directory for SHA
	 * hashing of the userPassword field.
	 */
	if (check_pass(passwd)) {
		syslog(LOG_WARNING,"%s uses a clear text password",dn);
		return 1;
	}

	/*
	 * No match!
	 */
	return 0;
}

/*
 * The main password checking routine. Uses the horse function to
 * perform the actual check, and give a nice welcome to the user if the
 * supplied password matches.
 */
int check_pass_ldap(char *passwd)
{
	char *s;

	if (check_pass_ldap_horse(passwd)) {
		lreply(230,"Welcome %s",cn?cn:dn);
		return 1;
	}

	return 0;
}

#endif /* USE_LDAP */

